/****************************************************/
/* GtpEngine                                        */
/* GTPで通信を行う                                  */
/* GTPコマンド全部を実装しているわけではありません  */
/* 実装しているGTPコマンドは下のコマンドリストを    */
/* 参照してください。                               */
/* GTPコマンドに関しては                            */
/* http://kamakama.s515.xrea.com/GTP.html           */
/* を参照してください。versionは２ですが、          */
/* GtpEngine.cではversion1ってことでよろしく        */
/****************************************************/

/***************************************************/
/* ヘッダーインクルード                            */
/***************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <ctype.h>

#include "gtp.h"
#include "gtpconv.h"
#include "board.h"
#include "rating.h"


/***************************************************/
/* コマンドリスト                                  */
/***************************************************/
/*
boardsize;
clear_board;
name;
protocol_version;
genmove;
play;
known_command;
list_commands;
quit;
komi;
get_komi;
final_score;
time_settings;
time_left;
version;
genmove_black
genmove_white
black
white
*/

/***************************************************/
/* グローバル変数                                  */
/***************************************************/
char* commandlist[COMMANDNUM];
void (*funclist[COMMANDNUM])();
char charset19x19[] = {' ', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T'};
int  intset19x19[]	= {0, 19, 18, 17, 16, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1};
char charset9x9[]	= {' ', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'J'};
int	 intset9x9[]	= {0, 9, 8, 7, 6, 5, 4, 3, 2, 1};
char input[INPUTBUF];
char input_copy[INPUTBUF];
char *next_token;

/***************************************************/
/* void GTP_main()                                 */
/* Gtp本体                                         */
/* goban : 碁盤本体                                */
/***************************************************/
void GTP_main(){
	int i;

	/***** 以下、変更部分 *****/
	clearboard();
	/***** 以上、変更部分 *****/

	GTP_setCommand();
	setvbuf(stdout, NULL, _IONBF, 0 );
	while(fgets(input, INPUTBUF, stdin) != NULL){
		char *command;
		int nocommand = 1;
		strcpy_s(input_copy, INPUTBUF, input);
		command = strtok_s(input, DELIM, &next_token);
		REMOVE_NEWLINE(command);
		for(i=0;i<COMMANDNUM;i++){
			if( !strcmp(command, commandlist[i]) ){
				(*funclist[i])();
				nocommand = 0;
				break;
			}
		}
		if(nocommand){
			GTP_response("? unknown command", TRUE);
		}
		fflush(stdin);
		fflush(stdout);
	}
}


/***************************************************/
/* void GTP_setcommand()                            */
/* GTPコマンドを設定する                           */
/***************************************************/
void GTP_setCommand(){
	commandlist[0]	= "boardsize";
	commandlist[1]	= "clear_board";
	commandlist[2]	= "name";
	commandlist[3]	= "protocol_version";
	commandlist[4]	= "genmove";
	commandlist[5]	= "play";
	commandlist[6]	= "known_command";
	commandlist[7]	= "list_commands";
	commandlist[8]	= "quit";
	commandlist[9]	= "komi";
	commandlist[10] = "get_komi";
	commandlist[11] = "final_score";
	commandlist[12] = "time_settings";
	commandlist[13] = "time_left";
	commandlist[14] = "version";
	commandlist[15] = "genmove_black";
	commandlist[16] = "genmove_white";
	commandlist[17] = "black";
	commandlist[18] = "white";			

	funclist[0]	 = GTP_boardsize;
	funclist[1]  = GTP_clearboard;
	funclist[2]	 = GTP_name;
	funclist[3]	 = GTP_protocolversion;
	funclist[4]  = GTP_genmove;
	funclist[5]  = GTP_play;
	funclist[6]  = GTP_knowncommand;
	funclist[7]  = GTP_listcommands;
	funclist[8]	 = GTP_quit;
	funclist[9]  = GTP_komi;
	funclist[10] = GTP_getkomi;
	funclist[11] = GTP_finalscore;
	funclist[12] = GTP_timesettings;
	funclist[13] = GTP_timeleft;
	funclist[14] = GTP_version;
	funclist[15] = GTP_genmove;
	funclist[16] = GTP_genmove;
	funclist[17] = GTP_play;
	funclist[18] = GTP_play;
}

/***************************************************/
/* void GTP_response(char *res, int success)       */
/* 標準出力へ文字列を出力                          */
/* res : 出力する文字列                            */
/***************************************************/
void GTP_response(char *res, int success){
	if(success)
		fprintf(stdout, "= ");
	else
		fprintf(stdout, "? ");
	fprintf(stdout, "%s\n\n", res);
	fflush(stdout);
}

/***************************************************/
/* void GTP_boardsize()                            */
/* boardsizeコマンドを処理                         */
/***************************************************/
void GTP_boardsize(){
	GTP_response("", clearboard());
}

/***************************************************/
/* void GTP_clearboard()                           */
/* clearboardコマンドを処理                        */
/***************************************************/
void GTP_clearboard(){
	GTP_response("", clearboard());
}

/***************************************************/
/* void GTP_name()                                 */
/* nameコマンドを処理                              */
/***************************************************/
void GTP_name(){
	GTP_response(PROGRAMNAME, TRUE);
}

/***************************************************/
/* void GTP_protocolversion()                      */
/* protocolversionコマンドを処理                   */
/***************************************************/
void GTP_protocolversion(){
	GTP_response(PROTOCOLVERSION, TRUE);
}

/***************************************************/
/* void GTP_genmove()                              */
/* genmoveコマンドを処理                           */
/***************************************************/
void GTP_genmove(){
	char *command;
	char c;
	char put[4];
	int color;
	int point, x, y;
	command = strtok_s(input_copy, DELIM, &next_token); 
	REMOVE_NEWLINE(command);
	if(!strcmp("genmove_black", command)){
		color = S_BLACK;
	}else if(!strcmp("genmove_white", command)){
		color = S_WHITE;
	}else{
		command = strtok_s(NULL, DELIM, &next_token);
		if(command == NULL){
			GTP_response("genmove color", TRUE);
			return;
		}
		REMOVE_NEWLINE(command);
		c = (char)tolower((int)command[0]);
		if(c == 'w')
			color = S_WHITE;
		else if(c == 'b')
			color = S_BLACK;
		else {
			GTP_response("genmove color", TRUE);
			return;
		}
	}

	/***** 以下、変更部分 *****/
	point = genmove(color);
	putstone(color, point);
	x = X(point)-(OB_SIZE-1);
	y = PURE_BOARD_SIZE - (Y(point)-(OB_SIZE-1)) + 1; 
	/***** 以上、変更部分 *****/

	if(point == PASS){
		GTP_response("PASS", TRUE);
		return;
	}else if(point == RESIGN){
		GTP_response("RESIGN", TRUE);
		return;
	}else if(!( x > 0 && x <= PURE_BOARD_SIZE && y > 0 && y <= PURE_BOARD_SIZE  )){
		GTP_response("", FALSE);
		return;
	}
	put[0] = charset19x19[x];
	if(y/10 == 0){
		put[1] = (char)('0'+y%10);
		put[2] = '\0';
	}else{
		put[1] = (char)('0'+y/10);
		put[2] = (char)('0'+y%10);
		put[3] = '\0';
	}
	GTP_response(put, TRUE);
}

/***************************************************/
/* void GTP_play()                                 */
/* playコマンドを処理                              */
/***************************************************/
void GTP_play(){
	char *command;
	char c;
	int x=0, y=0;
	char tmpy[2];
	int color;
	int i;
	command = strtok_s(input_copy, DELIM, &next_token);
	if(!strcmp("black", command)){
		color = S_BLACK;
	}else if(!strcmp("white", command)){
		color = S_WHITE;
	}else{
		command = strtok_s(NULL, DELIM, &next_token);
		if(command == NULL){
			GTP_response("play color point", FALSE);
			return ;
		}
		REMOVE_NEWLINE(command);
		c = (char)tolower((int)command[0]);
		if(c == 'w')
			color = S_WHITE;
		else
			color = S_BLACK;
	}
	command = strtok_s(NULL, DELIM, &next_token);
	REMOVE_NEWLINE(command);
	if(command == NULL){
		GTP_response("play color point", FALSE);
		return ;
	}else if(!strcmp(command, "pass") || !strcmp(command, "PASS")){
		x = 0;
		y = 0;
	}else{
		for(i=1;i<=PURE_BOARD_SIZE;i++){
			if(charset19x19[i] == command[0] || tolower((charset19x19[i])) == command[0]){
				x = i;
			}
		}
		tmpy[0] = command[1];
		tmpy[1] = command[2];
		y = PURE_BOARD_SIZE - atoi(tmpy) + 1;
	}

	/***** 以下、変更部分 *****/
	if( x==0 || y==0 ){
		putstone(color, PASS );
		GTP_response("PASS", TRUE);
		return;
	}else{
		if( x > 0 && x <= PURE_BOARD_SIZE && y > 0 && y <= PURE_BOARD_SIZE ){
			if( checklegal( color, POS(x+(OB_SIZE-1),y+(OB_SIZE-1)) ) == TRUE ){
				putstone(color, POS(x+(OB_SIZE-1),y+(OB_SIZE-1)) );
				GTP_response("", TRUE);
				return;
			}
		}
		GTP_response("", FALSE);
		return;
	}
	/***** 以上、変更部分 *****/
}

/***************************************************/
/* void GTP_knowncommand()                         */
/* knowncommandコマンドを処理                      */
/***************************************************/
void GTP_knowncommand(){
	int i;
	char *command;
	command = strtok_s(NULL, DELIM, &next_token);
	if(command == NULL){
		GTP_response("known_command command", FALSE);
		return;
	}
	REMOVE_NEWLINE(command);
	for(i=0;i<COMMANDNUM;i++){
		if( !strcmp(command, commandlist[i]) ){
			GTP_response("true", TRUE);
			return;
		}
	}
	GTP_response("false", FALSE);
}

/***************************************************/
/* void GTP_listcommands()                         */
/* listcommandsコマンドを処理                      */
/***************************************************/
void GTP_listcommands(){
	char list[512];
	unsigned int i, j, k;
	i = 0;
	list[i++] = '\n';
	for(j=0;j<COMMANDNUM;j++){
		for(k=0;k<strlen(commandlist[j]);k++){
			list[i++] = (commandlist[j])[k];
		}
		list[i++] = '\n';
	}
	list[i++] = '\0';
	GTP_response(list, TRUE);
}

/***************************************************/
/* void GTP_quit()                                 */
/* quitコマンドを処理                              */
/***************************************************/
void GTP_quit(){
	GTP_response("", fin());
	exit(0);
}

/***************************************************/
/* void GTP_komi()                                 */
/* komiコマンドを処理                              */
/***************************************************/
void GTP_komi(){
	char* komi;
	komi = strtok_s(NULL, DELIM, &next_token);
	if(komi != NULL){
		GTP_response("", TRUE);
	}
	else
		GTP_response("komi float", FALSE);
}

/***************************************************/
/* void GTP_getkomi()                              */
/* getkomiコマンドを処理                           */
/***************************************************/
void GTP_getkomi(){
	char buf[4];
	sprintf_s(buf, 4, "%.1f", KOMI);
	GTP_response(buf, TRUE);
}

/***************************************************/
/* void GTP_finalscore()                           */
/* finalscoreコマンドを処理                        */
/***************************************************/
void GTP_finalscore(){
	char buf[10];
	int score;
	score =  finalscore();
	if(score > 0)
		sprintf_s(buf, 10, "B+%d.5", score);
	else
		sprintf_s(buf, 10, "W+%d.5", abs(score));
	GTP_response(buf, TRUE);
}

/***************************************************/
/* void GTP_timesettings()                         */
/* timesettingsコマンドを処理                      */
/***************************************************/
void GTP_timesettings(){
	GTP_response("", TRUE);
}

/***************************************************/
/* void GTP_timeleft()                             */
/* timeleftコマンドを処理                          */
/***************************************************/
void GTP_timeleft(){
	GTP_response("", TRUE);
}

/***************************************************/
/* void GTP_version()                              */
/* versionコマンドを処理                           */
/***************************************************/
void GTP_version(){
	GTP_response(PROGRAMVERSION, TRUE);
}
